/**
 * Copyright (c) 2009 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/** @file
* @brief Example template project.
* @defgroup nrf_templates_example Example Template
*
*/

#include <stdbool.h>
#include <stdint.h>

#include "nrf.h"
#include "nordic_common.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_backend_rtt.h"
#include "nrf_delay.h"

#include "app_util.h"
#include "app_timer.h"
#include "nrf_drv_clock.h"
#include "nrf_stack_guard.h"
#include "stack_profiler.h"
#include "m_audio_gauges.h"

// Using Smart Remote 3 configuration files for easy reuse of Opus integration
#ifndef CONFIG_BOARD_NRF52832_CUSTOM
#define define CONFIG_BOARD_NRF52832_CUSTOM
#endif /* #ifndef CONFIG_BOARD_NRF52832_CUSTOM */
#include "sr3_config.h"
#include "sr3_config_nrf52832_custom.h"

#if CONFIG_CL_RTT_ENABLED
#if CONFIG_CL_RTT_INTERFACE == CONFIG_CL_INTERFACE_LOGGER
/**@brief Logger over RTT instance. */
NRF_LOG_BACKEND_RTT_DEF(m_rtt_log_backend);

#elif CONFIG_CL_RTT_INTERFACE == CONFIG_CL_INTERFACE_CLI
/**@brief Console over RTT instance. */
NRF_CLI_RTT_DEF(m_cli_rtt);
NRF_CLI_DEF(s_rtt_cli, "SR3-RTT> ", &m_cli_rtt.transport, '\n', CONFIG_LOG_RTT_QSIZE);
#else
#error "Unsupported RTT mode, please configure RTT as CLI or Logger backend"
#endif
#endif

#include "opus.h"

#define OPUS_MAX_FRAME_SIZE 3840

#if   (CONFIG_OPUS_MODE == CONFIG_OPUS_MODE_CELT)
# define OPUS_APPLICATION    OPUS_APPLICATION_RESTRICTED_LOWDELAY
# define OPUS_ENCODER_SIZE   7180
# define OPUS_DECODER_SIZE   9224
# define OPUS_MODE           "CELT"
# if (CONFIG_AUDIO_SAMPLING_FREQUENCY != 8000) && \
     (CONFIG_AUDIO_SAMPLING_FREQUENCY != 16000) && \
     (CONFIG_AUDIO_SAMPLING_FREQUENCY != 24000)
#  error "Selected sampling frequency is not supported by CELT codec."
# endif
# if (CONFIG_AUDIO_FRAME_SIZE_MS != 5) && \
     (CONFIG_AUDIO_FRAME_SIZE_MS != 10) && \
     (CONFIG_AUDIO_FRAME_SIZE_MS != 20)
#  error "Selected audio frame size is not supported by CELT codec."
# endif
#elif (CONFIG_OPUS_MODE == CONFIG_OPUS_MODE_SILK)
# define OPUS_APPLICATION    OPUS_APPLICATION_VOIP
# define OPUS_ENCODER_SIZE   10916
# define OPUS_MODE           "SILK"
# if (CONFIG_AUDIO_SAMPLING_FREQUENCY != 8000) && \
     (CONFIG_AUDIO_SAMPLING_FREQUENCY != 16000)
#  error "Selected sampling frequency is not supported by SILK codec."
# endif
# if (CONFIG_AUDIO_FRAME_SIZE_MS != 10) && \
     (CONFIG_AUDIO_FRAME_SIZE_MS != 20) && \
     (CONFIG_AUDIO_FRAME_SIZE_MS != 40) && \
     (CONFIG_AUDIO_FRAME_SIZE_MS != 60)
#  error "Selected audio frame size is not supported by SILK codec"
# endif
#else
# error "Unsupported OPUS Mode"
#endif

#if CONFIG_CLI_ENABLED
static bool                         m_opus_encoder_initialized;
static uint8_t                      m_opus_complexity = CONFIG_OPUS_COMPLEXITY;
static int32_t                      m_opus_bitrate    = ((CONFIG_OPUS_BITRATE != 0) ? CONFIG_OPUS_BITRATE : OPUS_AUTO);
static bool                         m_opus_vbr        = ((CONFIG_OPUS_BITRATE == 0) || (CONFIG_OPUS_VBR_ENABLED != 0));
#else /* !CONFIG_CLI_ENABLED */
# define                            m_opus_complexity   CONFIG_OPUS_COMPLEXITY
# define                            m_opus_bitrate      ((CONFIG_OPUS_BITRATE != 0) ? CONFIG_OPUS_BITRATE : OPUS_AUTO)
# define                            m_opus_vbr          ((CONFIG_OPUS_BITRATE == 0) || (CONFIG_OPUS_VBR_ENABLED != 0))
#endif /* CONFIG_CLI_ENABLED */

__ALIGN(4) static uint8_t           m_opus_encoder[OPUS_ENCODER_SIZE];
static OpusEncoder * const          m_opus_encoder_state = (OpusEncoder *)m_opus_encoder;

__ALIGN(4) static uint8_t           m_opus_decoder[OPUS_DECODER_SIZE];
static OpusDecoder * const          m_opus_decoder_state = (OpusDecoder *)m_opus_decoder;

static m_audio_bitrate_gauge_t  m_bitrate_gauge;
static m_audio_cpu_gauge_t      m_encoding_cpu_gauge;
static m_audio_cpu_gauge_t      m_decoding_cpu_gauge;

extern const uint8_t sample1[215110];

static void drv_audio_codec_log_config(const char *action)
{
    if (m_opus_bitrate == OPUS_AUTO)
    {
        NRF_LOG_INFO("OPUS/" OPUS_MODE " Codec %s (mode: VBR, complexity: %u, frame: %u ms).",
                     action,
                     m_opus_complexity,
                     CONFIG_AUDIO_FRAME_SIZE_MS);
    }
    else
    {
       NRF_LOG_INFO("OPUS/" OPUS_MODE " Codec %s (mode: %s %u kbit/s, complexity: %u, frame: %u ms)",
                    action,
                    (m_opus_vbr) ? "CVBR" : "CBR",
                    m_opus_bitrate / 1000,
                    m_opus_complexity,
                    CONFIG_AUDIO_FRAME_SIZE_MS);
    }
}

static void opus_encoder_configure(void)
{
    ASSERT(opus_encoder_get_size(1) == sizeof(m_opus_encoder));

    NRF_LOG_INFO("opus_encoder_get_size(1) = %d", opus_encoder_get_size(1));

    APP_ERROR_CHECK_BOOL(opus_encoder_init(m_opus_encoder_state, CONFIG_AUDIO_SAMPLING_FREQUENCY, 1, OPUS_APPLICATION) == OPUS_OK);

    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_BITRATE(m_opus_bitrate))                      == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_VBR(m_opus_vbr))                              == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_VBR_CONSTRAINT((m_opus_bitrate != OPUS_AUTO)))== OPUS_OK);

    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_COMPLEXITY(m_opus_complexity))                == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_SIGNAL(OPUS_SIGNAL_VOICE))                    == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_LSB_DEPTH(16))                                == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_DTX(0))                                       == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_INBAND_FEC(0))                                == OPUS_OK);
    APP_ERROR_CHECK_BOOL(opus_encoder_ctl(m_opus_encoder_state, OPUS_SET_PACKET_LOSS_PERC(0))                          == OPUS_OK);
}

static void opus_decoder_configure(void)
{
    ASSERT(opus_decoder_get_size(1) == sizeof(m_opus_decoder));
    
    NRF_LOG_INFO("opus_decoder_get_size(1) = %d", opus_decoder_get_size(1));

    APP_ERROR_CHECK_BOOL(opus_decoder_init(m_opus_decoder_state, CONFIG_AUDIO_SAMPLING_FREQUENCY, 1) == OPUS_OK);
}

static void opus_run_loopback_test(void)
{
    static int16_t frame_buf_uncompressed[CONFIG_AUDIO_FRAME_SIZE_SAMPLES];
    static uint8_t frame_buf_compressed[CONFIG_AUDIO_FRAME_SIZE_BYTES];
    size_t         num_frames_in_sample;
    opus_int32     compressed_frame_size;
    opus_int32     compressed_frame_size_max;
    int            decompressed_frame_size;

#if CONFIG_STACK_PROFILER_ENABLED
    uint32_t        encoder_mem_usage;
    uint32_t        encoder_mem_usage_max;
    uint32_t        decoder_mem_usage;
    uint32_t        decoder_mem_usage_max;
#endif

    m_audio_bitrate_gauge_reset(&m_bitrate_gauge);
    m_audio_cpu_gauge_reset(&m_encoding_cpu_gauge);
    m_audio_cpu_gauge_reset(&m_decoding_cpu_gauge);

    num_frames_in_sample = (sizeof(sample1) / sizeof(int16_t)) / CONFIG_AUDIO_FRAME_SIZE_SAMPLES;

    NRF_LOG_DEBUG("num_frames_in_sample:  %d", num_frames_in_sample);

    compressed_frame_size_max = 0;

#if CONFIG_STACK_PROFILER_ENABLED
    encoder_mem_usage_max = 0;
    decoder_mem_usage_max = 0;
#endif

    for (int i = 0; i < num_frames_in_sample; ++i)
    {
        // Copy frame from sample
        memcpy(frame_buf_uncompressed, &((int16_t*)sample1)[i * CONFIG_AUDIO_FRAME_SIZE_SAMPLES], CONFIG_AUDIO_FRAME_SIZE_SAMPLES * sizeof(int16_t));

#if CONFIG_STACK_PROFILER_ENABLED
    // Initialize stack profiler.
    stack_profiler_init();
#endif
        // Compress frame
        m_audio_measure_cpu_usage_start(&m_encoding_cpu_gauge);
        compressed_frame_size = opus_encode(m_opus_encoder_state, 
                                            frame_buf_uncompressed, 
                                            CONFIG_AUDIO_FRAME_SIZE_SAMPLES, 
                                            frame_buf_compressed,
                                            sizeof(frame_buf_compressed));
        m_audio_measure_cpu_usage_end(&m_encoding_cpu_gauge);

        APP_ERROR_CHECK_BOOL((compressed_frame_size > 0) && (compressed_frame_size <= OPUS_MAX_FRAME_SIZE));

        m_audio_measure_bitrate(&m_bitrate_gauge, compressed_frame_size);

        if (compressed_frame_size > compressed_frame_size_max)
        {
            compressed_frame_size_max = compressed_frame_size;
        }

#if CONFIG_STACK_PROFILER_ENABLED
        encoder_mem_usage = stack_profiler_get_max_stack_usage();
        if (encoder_mem_usage > encoder_mem_usage_max)
        {
            encoder_mem_usage_max = encoder_mem_usage;
        }

        // Initialize stack profiler.
        stack_profiler_init();
#endif

        // Decompress frame
        m_audio_measure_cpu_usage_start(&m_decoding_cpu_gauge);
        decompressed_frame_size = opus_decode(m_opus_decoder_state, frame_buf_compressed, compressed_frame_size, frame_buf_uncompressed, CONFIG_AUDIO_FRAME_SIZE_SAMPLES, 0);
        m_audio_measure_cpu_usage_end(&m_decoding_cpu_gauge);
        APP_ERROR_CHECK_BOOL(decompressed_frame_size == CONFIG_AUDIO_FRAME_SIZE_SAMPLES);

#if CONFIG_STACK_PROFILER_ENABLED
        decoder_mem_usage = stack_profiler_get_max_stack_usage();
        if (decoder_mem_usage > decoder_mem_usage_max)
        {
            decoder_mem_usage_max = decoder_mem_usage;
        }
#endif
    }

    m_audio_bitrate_gauge_log(&m_bitrate_gauge, "\t- Bitrate");
    m_audio_cpu_gauge_log(&m_encoding_cpu_gauge, "\t- Encoder");
    m_audio_cpu_gauge_log(&m_decoding_cpu_gauge, "\t- Decoder");
    NRF_LOG_INFO("Max compressed frame size: %u bytes", compressed_frame_size_max);

#if CONFIG_STACK_PROFILER_ENABLED
    NRF_LOG_INFO("Maximum encoder stack usage: %u%% (%u out of %u bytes)",
              100ul * encoder_mem_usage_max / REAL_STACK_SIZE,
              encoder_mem_usage_max,
              REAL_STACK_SIZE);

    NRF_LOG_INFO("Maximum decoder stack usage: %u%% (%u out of %u bytes)",
              100ul * decoder_mem_usage_max / REAL_STACK_SIZE,
              decoder_mem_usage_max,
              REAL_STACK_SIZE);
#endif
}

/**
 * @brief Function for application main entry.
 */
int main(void)
{
    // Enable I-Cache.
    NRF_NVMC->ICACHECNF |= NVMC_ICACHECNF_CACHEEN_Msk;

    // Initialize Clock Management Driver.
    APP_ERROR_CHECK(nrf_drv_clock_init());

    // Start LFCLK.
    nrf_drv_clock_lfclk_request(NULL);

    app_timer_init();
    app_timer_resume();

#if CONFIG_LOG_ENABLED
    // Initialize SDK Logger.
    APP_ERROR_CHECK(NRF_LOG_INIT(app_timer_cnt_get));
#endif

#if CONFIG_CL_RTT_ENABLED
    /*
     * Start RTT console if debugger is connected. RTT console may be one of the logger backends
     * so it needs to be attached to logger as soon as possible.
     */
    if (CoreDebug->DHCSR & CoreDebug_DHCSR_C_DEBUGEN_Msk)
    {
#if CONFIG_CL_RTT_INTERFACE == CONFIG_CL_INTERFACE_LOGGER
        nrf_log_backend_rtt_init();
        int32_t rtt_backend_id = nrf_log_backend_add(&m_rtt_log_backend.backend,
                                                     (nrf_log_severity_t)CONFIG_LOG_RTT_SEVERITY);
        if (rtt_backend_id < 0)
        {
            ASSERT(false);
        }
        nrf_log_backend_enable(&m_rtt_log_backend.backend);
#endif

#if CONFIG_CL_RTT_INTERFACE == CONFIG_CL_INTERFACE_CLI
        APP_ERROR_CHECK(nrf_cli_init(&s_rtt_cli,
                                     NULL,
                                     true,
                                     CONFIG_CLI_RTT_LOG_BACKEND_ENABLED,
                                     (nrf_log_severity_t)(CONFIG_LOG_RTT_SEVERITY)));
#endif
    }
#endif

#if CONFIG_STACK_PROFILER_ENABLED
    // Initialize stack profiler.
    stack_profiler_init();
#endif

    if (REAL_STACK_SIZE < (CONFIG_STACK_SIZE - CONFIG_STACK_SIZE_GUARD))
    {
        NRF_LOG_WARNING("Stack size is lower than recommended (%d vs %d bytes)", REAL_STACK_SIZE, (CONFIG_STACK_SIZE - CONFIG_STACK_SIZE_GUARD));
    }
    else
    {
        NRF_LOG_INFO("Usable stack size: %d bytes", (CONFIG_STACK_SIZE - CONFIG_STACK_SIZE_GUARD));
    }

    opus_encoder_configure();
    opus_decoder_configure();

    drv_audio_codec_log_config("initialized");

    opus_run_loopback_test();

    while (true)
    {
        __WFE();
    }
}
/** @} */
